import os
import sys
import platform
import json
import types

__version__ = "1.0.1"
sys.path.insert(0, os.path.dirname(__file__))

if platform.system() == "Windows":
    mypath = os.path.dirname(os.path.realpath(__file__))
    if mypath not in sys.path:
        sys.path.insert(0, mypath)
    os.add_dll_directory(mypath)

from .cpppart import cpppart as base
if base.use_pyxyz:
    from .pyxyz import Confpool, MolProxy
# from .pyutils import pyutils
import networkx as nx
import pyutils

DEG2RAD = 0.0174532925199432957692
RAD2DEG = 1 / DEG2RAD
H2KC = 627.509474063
KC2H = 1 / H2KC


# Factory of Molecule objects
def Molecule(sdf=None,
             graph_data=None,
             request_free=None,
             require_best_sequence=False,
             clear_feed=True):
    assert (sdf is not None) or (
        graph_data is not None
    ), "Either 'sdf' or 'graph_data' must be passed as keyword-args"
    assert (sdf is None) or (
        graph_data is None
    ), "'sdf' or 'graph_data' should not be specified simultaneously"

    # Clear feed for new molecule
    if clear_feed:
        base.clear_status_feed()

    m = base.Molecule()

    if request_free is not None:
        request_free = [(i - 1, j - 1) for i, j in request_free]
        m.set_requested_dofs(request_free)

    if require_best_sequence:
        m.require_best_sequence()

    if sdf is not None:
        assert isinstance(sdf,
                          str), "sdf must be an str object (name of your SDF)"
        m.sdf_constructor(sdf, nx, pyutils)
    else:  # elif graph_data is not None
        assert isinstance(
            graph_data,
            dict), "graph_data must be a dict (see README for details)"
        assert set(graph_data.keys()) == {
            'graph', 'fixed_dihedrals'
        }, "graph_data keys must be precisely 'graph' and 'fixed_dihedrals'"
        m.graph_constructor(graph_data, nx, pyutils)
    return m


def get_one_seed():
    return int.from_bytes(os.urandom(3), byteorder="big")


def create_seed_list(size):
    unique_set = set()
    result = []
    while len(result) < size:
        list_element = get_one_seed()
        if list_element not in unique_set:
            unique_set.add(list_element)
            result.append(list_element)
    return result


MCR_FLAGS_DESCRIPTIONS = {
    'total': {
        'ntries': "Number of sampling attempts in total"
    },
    'good': {
        'nsucc':
        "Successfull iterations of MCR with postoptimization (if it's enabled)",
    },
    'bad': {
        'nfail':
        "Cases of TLC unable to provide correct solutions (only incorrect ones)",
        'nzero': "Cases of zero IK solutions",
        'ngeom': "MCR sampling attempts didn't pass geometry validation",
        'nolap': "MCR sampling attempts didn't pass overlap checks",
        'npostopt_fail':
        "Crashed postoptimizations (BFGS unable to make a step, etc.)",
        'npostopt_more_steps':
        "Postoptimization unable to pass the overlap and validity requirements",
        'ndihedral_filtering':
        "Discarded due to violated configuration rules for dihedrals",
        'nrmsd_duplicate': "Discarded by RMSD filter",
    }
}


def mcr_result_to_list(result_data: dict) -> dict:
    result_descriptions = {}
    for section_name, section_cases in MCR_FLAGS_DESCRIPTIONS.items():
        result_descriptions[section_name] = {}
        for key, description in section_cases.items():
            if key in result_data:
                result_descriptions[section_name][description] = result_data[
                    key]
    return result_descriptions


def run_confsearch(
    mol,
    pool=None,
    rmsd_settings=None,
    postopt_settings=[],
    geometry_validation={},
    nthreads=1,
    max_conformers=None,
    max_tries=None,
    timelimit=None,
    filter=None,
    clear_feed=True,
):
    max_conformers = max_conformers if max_conformers is not None else -1
    max_tries = max_tries if max_tries is not None else -1
    timelimit = timelimit if timelimit is not None else -1

    # Check if all requrements are supported by the used library
    # Concerning 'nthreads'
    assert isinstance(nthreads, int), "'nthreads' must be interger"
    if nthreads == 1:
        assert base.use_mcr_singlethreaded, "The default single-threaded Monte-Carlo is not supported by the current build of Ringo"
    elif nthreads > 1:
        assert base.use_mcr_parallel, "Parallel Monte-Carlo is not supported by the current build of Ringo"
    else:
        raise RuntimeError("'nthreads' must be >= 1")
    if not base.use_postoptimization:
        assert (len(postopt_settings) == 0) or \
            (postopt_settings == [{'enabled': False}, {'enabled': False}]), \
            "Postoptimization is not supported by this build"
    # Concerning pool object and RMSD filtering settings
    if (pool is not None) or (rmsd_settings
                              is not None) or (filter
                                               is not None) or base.use_pyxyz:
        assert base.use_pyxyz, "Either 'pool', 'rmsd_settings' or 'filter' were specified but Pyxyz is not included in the current build"
        assert pool is not None, "'pool' for conformer storage must be provided"
        # Set default rmsd_settings
        if rmsd_settings is None or rmsd_settings == 'default':
            # Using default settings for RMSD filtering
            rmsd_settings = {
                'isomorphisms': {
                    'ignore_elements': ['HCarbon'],
                },
                'rmsd': {
                    'threshold': 0.2,
                    'mirror_match': True,
                }
            }
        assert isinstance(
            pool, base.Confpool), "Pool must be an instance of pyxyz.Confpool"
        assert isinstance(rmsd_settings,
                          dict), "'rmsd_settings' must be a dict"
        assert filter is None or isinstance(
            filter, types.FunctionType) or isinstance(
                filter, dict), "'filter' must be a function or dict"
    # Other execution options
    assert isinstance(max_conformers, int) and (
        max_conformers >= 1
        or max_conformers == -1), "Incorrect value of 'max_conformers' keyword"
    assert isinstance(
        max_tries, int) and (max_tries >= 1 or max_tries
                             == -1), "Incorrect value of 'max_tries' keyword"
    assert isinstance(
        timelimit, int) and (timelimit >= 1 or timelimit
                             == -1), "Incorrect value of 'timelimit' keyword"
    if max_conformers == -1 and max_tries == -1 and timelimit == -1:
        raise RuntimeError(
            "At least one keyword 'max_conformers', 'max_tries' or 'timelimit' must be provided. "
            "Otherwise, Monte-Carlo run has no condition for termination.")
    # Clear feed for confsearch run
    if clear_feed:
        base.clear_status_feed()

    # Assemble argument list for Monte-Carlo call
    arg_list = [mol]
    if base.use_pyxyz:
        # Options for conformer storage and filtering
        arg_list += [pool, filter, rmsd_settings]
    if base.use_postoptimization:
        arg_list.append(postopt_settings)
    if base.use_geometry_validation:
        arg_list.append(geometry_validation)
    if nthreads == 1:
        # Option for single-threaded runs
        arg_list.append(get_one_seed())
    else:
        # Options for parallel runs
        arg_list += [nthreads, create_seed_list(nthreads)]
    # Options for termination criteria
    arg_list += [max_conformers, max_tries, timelimit]
    # Non-C++ modules
    # if nthreads > 1:
    arg_list += [nx, pyutils]

    # Execute Monte-Carlo
    if nthreads == 1:
        res = base.mcr_confsearch(*arg_list)
    else:
        res = base.mcr_confsearch_parallel(*arg_list)

    return res


def cleanup():
    seqcache_path = os.path.join(os.getcwd(), pyutils.CACHE_FILE)
    if os.path.isfile(seqcache_path):
        os.remove(seqcache_path)
        # print(f"Removing cache file {seqcache_path}")


# Work with vdw radii controls
build_flags = base.build_flags
if base.use_overlap_detection:
    get_vdw_radii = base.get_vdw_radii
    set_vdw_radii = base.set_vdw_radii
    set_radius_multiplier = base.set_radius_multiplier

# Work with Ringo status feed
WARNING_CODES = base.warning_codes
for warning_code, warning_line in base.warning_codes.items():
    globals(
    )[warning_code] = warning_line  # Declares str variables IK_NOT_APPLIED, SUBOPTIMAL_SOLN_SEQ, UNMET_DOF_REQUEST, etc.
add_message_to_feed = base.add_message_to_feed
clear_status_feed = base.clear_status_feed


def get_status_feed(important_only=True):
    json_data = base.get_status_feed()
    parsed_data = [json.loads(item) for item in json_data]
    for item in parsed_data:
        item['important'] = '[important]' in item['subject']
        item['subject'] = item['subject'].replace('[important]', '')
        if len(item['atoms']) > 0:
            item['atoms'] = sorted([idx + 1 for idx in item['atoms']])
        else:
            del item['atoms']

    if important_only:
        return [item for item in parsed_data if item['important']]
    else:
        return parsed_data


# Summarising statistics for the molecule
def get_molecule_statistics(m):
    graph = m.molgraph_access()
    symbols = m.get_symbols()

    composition = set(symbols[atom] for atom in graph.nodes)
    composition = {element: 0 for element in composition}
    for atom in graph.nodes:
        composition[symbols[atom]] += 1

    temp_graph = nx.Graph()
    temp_graph.add_edges_from(graph.edges)
    molgraph_bridges = list(nx.bridges(temp_graph))
    temp_graph.remove_edges_from(molgraph_bridges)
    lone_nodes = []
    for node in temp_graph.nodes:
        if len(list(temp_graph.neighbors(node))) == 0:
            lone_nodes.append(node)
    temp_graph.remove_nodes_from(lone_nodes)
    num_cyclic_parts = len([x for x in nx.connected_components(temp_graph)])
    all_ring_atoms = list(temp_graph.nodes)

    mcb = [set(c) for c in nx.minimum_cycle_basis(graph)]
    num_cyclic_rotatable_bonds = 0
    num_rotatable_bonds = 0
    for vxA, vxB in graph.edges:
        if len(list(graph.neighbors(vxA))) > 1 and len(
                list(graph.neighbors(vxB))) > 1:
            num_rotatable_bonds += 1
            for ring_atoms in mcb:
                if vxA in ring_atoms and vxB in ring_atoms:
                    num_cyclic_rotatable_bonds += 1
                    break

    num_cyclic_dofs = 0
    dof_list, _ = m.get_ps()
    for sideA, atA, atB, sideB in dof_list:
        if atA in all_ring_atoms and atB in all_ring_atoms and \
                (atA, atB) not in molgraph_bridges and (atB, atA) not in molgraph_bridges:
            num_cyclic_dofs += 1

    result = {
        'composition':
        composition,
        'num_atoms':
        graph.number_of_nodes(),
        'num_heavy_atoms':
        len([atom for atom in graph.nodes if symbols[atom] != 'H']),
        'num_bonds':
        graph.number_of_edges(),
        'num_rotatable_bonds':
        num_rotatable_bonds,
        'num_cyclic_rotatable_bonds':
        num_cyclic_rotatable_bonds,
        'largest_macrocycle_size':
        max([len(item) for item in mcb]) if len(mcb) > 0 else 0,
        'n_flexible_rings':
        m.get_num_flexible_rings(),
        'n_rigid_rings':
        m.get_num_rigid_rings(),
        'num_dofs':
        len(dof_list),
        'num_cyclic_dofs':
        num_cyclic_dofs,
        'cyclomatic_number':
        graph.number_of_edges() - graph.number_of_nodes() + 1,
        'num_cyclic_parts':
        num_cyclic_parts,
    }
    return result



import networkx as nx
import numpy as np
from copy import deepcopy
from typing import Tuple, List

from .base_load import base


class Molecule:

    def __init__(self, start_obj: base.MolProxy) -> None:
        if not isinstance(start_obj, base.MolProxy):
            raise RuntimeError("Molecule should be initialized from MolProxy")

        self.G = nx.Graph(start_obj.molgraph)
        xyz = start_obj.xyz
        for i in self.G.nodes:
            self.G.nodes[i]['xyz'] = xyz[i, :]
        self.G = nx.convert_node_labels_to_integers(self.G)

    def set_orientation(self, U: np.ndarray, new_center: np.ndarray) -> None:
        n_atoms = self.G.number_of_nodes()
        my_center = np.zeros(3)
        for node in self.G.nodes:
            my_center += self.G.nodes[node]['xyz']
        my_center /= n_atoms
        for node in self.G.nodes:
            self.G.nodes[node]['xyz'] -= my_center
            self.G.nodes[node]['xyz'] = self.G.nodes[node]['xyz'] @ U
            self.G.nodes[node]['xyz'] += new_center

    def __add__(self, other: 'Molecule') -> 'Molecule':
        res = deepcopy(other)
        n_reserve = self.G.number_of_nodes()
        mapping = {node: n_reserve + node for node in res.G.nodes}
        res.G = nx.relabel_nodes(res.G, mapping)
        res.G = nx.compose(res.G, self.G)
        return res

    def save_sdf(self, fname: str) -> None:
        lines = ["", "", ""]
        lines.append("%3d%3d  0  0  0  0  0  0  0  0999 V2000" %
                     (self.G.number_of_nodes(), self.G.number_of_edges()))
        for i in range(self.G.number_of_nodes()):
            lines.append(
                "%10.4f%10.4f%10.4f%3s  0  0  0  0  0  0  0  0  0  0  0  0" %
                (self.G.nodes[i]['xyz'][0], self.G.nodes[i]['xyz'][1],
                 self.G.nodes[i]['xyz'][2], self.G.nodes[i]['symbol']))

        for edge in self.G.edges:
            lines.append("%3s%3s%3s  0" % (edge[0] + 1, edge[1] + 1, 1))
        lines.append("M  END\n")

        with open(fname, "w") as f:
            f.write("\n".join(lines))

    def as_xyz(self) -> Tuple[List[np.ndarray], List[str]]:
        xyzs = []
        syms = []
        for atom in range(self.G.number_of_nodes()):
            xyzs.append(self.G.nodes[atom]['xyz'])
            syms.append(self.G.nodes[atom]['symbol'])
        return xyzs, syms



from .tests import RINGO_TESTS, base
from typing import Union


class FailedTestError(Exception):

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


def run_tests(print_result=True) -> Union[None, str]:
    resA = run_ringo_tests(print_result=print_result)
    resB = run_pyxyz_tests(print_result=print_result)
    if not print_result:
        return resA + '\n' + resB


def run_pyxyz_tests(print_result=True) -> Union[None, str]:
    if base.use_pyxyz:
        import ringo.pyxyz.test
        return ringo.pyxyz.test.run_tests(print_result=print_result)
    else:
        print("Skipping PyXYZ tests because it won't be used")


def run_ringo_tests(print_result=True) -> Union[None, str]:
    print("Running tests...")
    failed_tests = {}
    for test_name, test_call in RINGO_TESTS.items():
        success, message = test_call()
        if not success:
            failed_tests[test_name] = message

    if len(failed_tests) == 0:
        message_lines = ['All tests completed successfully:']
        for i, test_name in enumerate(RINGO_TESTS.keys(), start=1):
            message_lines.append(f"✔ {i}) {test_name}")
        message = '\n'.join(message_lines)
        if print_result:
            print(message)
            return None
        else:
            return message

    num_failed_tests = len(failed_tests)
    num_tests_total = len(RINGO_TESTS)
    error_lines = [
        f"{num_failed_tests} test(s) out of {num_tests_total} had failed:"
    ]
    for i, (test_name, message) in enumerate(failed_tests.items(), start=1):
        error_lines.append(f"{i}) {test_name}: {message}")
    raise FailedTestError('\n'.join(error_lines))



import random
from functools import wraps
from typing import Callable

import numpy as np

from .utils import (
    match_lists,
    to_abs,
    DOUBLE_THR,
)

from .. import (
    Confpool,
    base,
    H2KC,
    KC2H,
    Molecule,
)

MolProxy = base.MolProxy

PYXYZ_TESTS = {}


def test(raw_f: Callable) -> Callable:

    @wraps(raw_f)
    def f():
        try:
            raw_f()
            success = True
            message = ''
        except Exception as e:
            success = False
            message = repr(e)
        return (success, message)

    PYXYZ_TESTS[f.__name__] = f
    return f


@test
def load_hydrogen():
    p = Confpool()
    p.include_from_file(to_abs("hydrogen_atom.xyz"))
    assert p.size == len(p) == 1, "Incorrect number of conformers"
    assert p.natoms == 1, "Incorrect number of conformers"
    assert p.atom_symbols == ['H'], "Incorrect number of conformers"
    m = p[0]
    assert match_lists(
        m.xyz.tolist(),
        [[0.0, 0.0, 0.0]]), (f"{m.xyz.tolist()} vs. {[0.0,0.0,0.0]}")
    assert m.idx == 0
    assert m.natoms == 1
    assert m.descr == "Energy = -1488.228"


@test
def empty_filter():
    p = Confpool()
    p.include_from_file(to_abs("aminoacid_single.xyz"))
    p.filter(lambda m: m.l(1, 19) < 2.0)
    assert p.size == 0, "Incorrect number of conformers"


def _load_ensemble():
    p = Confpool()
    p.include_from_file(to_abs("aminoacid_ensemble.xyz"))
    p["Energy"] = lambda m: float(m.descr.strip())
    p.descr = lambda m: "Conf #{}; Energy = {:9f} a.u.".format(
        m.idx + 1, m["Energy"])
    return p


@test
def description_edit():
    p = _load_ensemble()
    p.float_from_descr("Energy_check", 1)
    assert match_lists(
        p["Energy"],
        p["Energy_check"]), f'{p["Energy"]} vs.\n {p["Energy_check"]}'


@test
def filtering():
    p = _load_ensemble()
    p.upper_cutoff("Energy", 5.0 * KC2H)
    assert all(m["Energy"] < 5.0 * KC2H for m in p)
    p.filter(lambda m: m.l(1, 19) < 2.0)
    assert all(m.l(1, 19) < 2.0 for m in p)


@test
def sorting():
    p = _load_ensemble()
    p.sort("Energy")
    energies = p['Energy']
    assert all(i < j
               for i, j in zip(energies, energies[1:])), "Basic sort failed"
    p.sort("Energy", ascending=False)
    energies = p['Energy']
    assert all(
        i > j
        for i, j in zip(energies, energies[1:])), "Descending sort failed"


@test
def attrs():
    p = _load_ensemble()
    xyz = p[0].xyz
    assert all(m.idx == i for i, m in enumerate(p)), "'idx' attr failed"
    assert xyz.shape[1] == 3
    natoms_ref = xyz.shape[0]
    assert p.natoms == natoms_ref, (
        f"Confpool natoms fail: {p.natoms} (ref={natoms_ref})")
    assert p[0].natoms == natoms_ref, (
        f"MolProxy natoms fail: {p[0].natoms} (ref={natoms_ref})")
    assert all(m.descr.startswith(f"Conf #{i+1}")
               for i, m in enumerate(p)), ("'descr' attr failed")


@test
def rmsd():
    p = Confpool()
    p.include_from_file(to_abs("crest_conformersA.xyz"))
    p.include_from_file(to_abs("crest_conformersB.xyz"))
    assert len(p) == 365
    p.generate_connectivity(0,
                            mult=1.3,
                            sdf_name=to_abs('test_topology.sdf'),
                            ignore_elements=['HCarbon'],
                            add_bonds=[[13, 23]])
    p.generate_isomorphisms()
    rmsd_matrix: np.ndarray = p.get_rmsd_matrix()
    assert rmsd_matrix.shape[0] == rmsd_matrix.shape[
        1], "RMSD matrix shape fail"
    assert all((rmsd_matrix[i, i] <= DOUBLE_THR and rmsd_matrix[i, i] >= 0.0)
               for i in range(rmsd_matrix.shape[0])), (
                   "RMSD matrix nonzero diagonal fail")
    assert all(
        abs(rmsd_matrix[i, j] - rmsd_matrix[j, i]) <= DOUBLE_THR
        for i in range(rmsd_matrix.shape[0])
        for j in range(i)), ("RMSD matrix symmetry fail")

    i, j = random.randint(0, rmsd_matrix.shape[0] - 1), random.randint(
        0, rmsd_matrix.shape[0] - 1)
    manual_rmsd = p[i].rmsd(p[j])[0]
    assert abs(manual_rmsd - rmsd_matrix[i, j]) < DOUBLE_THR, (
        f"Manual {manual_rmsd} vs. matrix {rmsd_matrix[i, j]} RMSD for i={i}, j={j}"
    )
    p.rmsd_filter(0.3, rmsd_matrix=rmsd_matrix)

    pp = Confpool()
    pp.include_from_file(to_abs("crest_conformersA.xyz"))
    pp.include_from_file(to_abs("crest_conformersB.xyz"))
    pp.generate_connectivity(0,
                             mult=1.3,
                             ignore_elements=['HCarbon'],
                             add_bonds=[[13, 23]])
    pp.generate_isomorphisms()
    pp.rmsd_filter(0.3)
    assert len(p) == len(pp), (
        f"Filtering cross check failed: {len(p)} vs. {len(pp)}")


@test
def rmsd_pair_iter():
    p = _load_ensemble()
    p.generate_connectivity(0, mult=1.3, ignore_elements=['HCarbon'])
    p.generate_isomorphisms()
    rmsd_matrix = p.get_rmsd_matrix()

    min_rmsd = 0.2
    max_rmsd = 0.25
    check_entry_count = sum(
        1 for i in range(len(p)) for j in range(i)
        if rmsd_matrix[i, j] >= min_rmsd and rmsd_matrix[i, j] <= max_rmsd)

    entry_count = sum(1
                      for _ in p.rmsd_fromto(min_rmsd, max_rmsd, rmsd_matrix))

    ascending_rmsd = [
        rmsd for _, __, rmsd in p.rmsd_fromto(min_rmsd, max_rmsd, rmsd_matrix)
    ]
    descending_rmsd = [
        rmsd for _, __, rmsd in p.rmsd_fromto(max_rmsd, min_rmsd, rmsd_matrix)
    ]

    assert ascending_rmsd == sorted(ascending_rmsd)
    assert descending_rmsd == sorted(ascending_rmsd, key=lambda x: -x)

    assert entry_count == check_entry_count, f"actual {entry_count} vs. check {check_entry_count}"


@test
def molgraph():

    def check(p):
        confA = Molecule(p[0])
        confB = Molecule(p[1])
        assert (confA.G.number_of_edges()
                != 0) and (confA.G.number_of_nodes()
                           != 0), ("Num nodes or Num edges == 0")
        a_data = (confA.G.number_of_edges(), confA.G.number_of_nodes())
        b_data = (confB.G.number_of_edges(), confB.G.number_of_nodes())
        assert a_data == b_data, f"Compare fail: {a_data} vs. {b_data}"
        gsum = confA + confB
        sum_data = (gsum.G.number_of_edges(), gsum.G.number_of_nodes())
        assert sum_data == (a_data[0] * 2, a_data[1] *
                            2), (f"Sum fail. {sum_data} vs. {a_data}")
        gsum.save_sdf(to_abs('test_sum.sdf'))

    p = _load_ensemble()
    p.generate_connectivity(0, mult=1.3)
    check(p)
    p = _load_ensemble()
    p.generate_connectivity(0, mult=1.3, ignore_elements=['HCarbon'])
    check(p)


@test
def geom_analysis():
    p = Confpool()
    p.include_from_file(to_abs("geom_analysis.xyz"))
    m = p[0]
    assert match_lists([m.l(1, 2), m.l(2, 3), m.l(3, 4)],
                       [0.96, 1.67, 2.0]), "Length check failed"
    assert match_lists([m.v(1, 2, 3), m.v(2, 3, 4)],
                       [120.0, 140.0]), "Vangle check failed"
    assert abs(m.z(1, 2, 3, 4) + 50.0) < DOUBLE_THR, (
        f"Dihedral check failed: {m.z(1, 2, 3, 4)} vs. -50.0")

    init_coords = [[0.0, 0.0, 0.0], [1.0, 2.0, 3.0], [10.0, 20.0, 10.0],
                   [-1.0, -1.0, -1.0]]
    p.include_from_xyz(np.matrix(init_coords), "Test descr")
    m = p[2]
    xyz = m.xyz
    assert match_lists(xyz.tolist(), init_coords)
    xyz[0] = [-5.0, -5.0, -5.0]
    p.include_from_xyz(xyz, "Test descr2")
    assert abs(p[3].l(1, 2) - 12.206555615733702) < DOUBLE_THR
    p.generate_connectivity(0,
                            mult=1.3,
                            sdf_name=to_abs('geom_topology.sdf'),
                            ignore_elements=['HCarbon'])
    p.generate_isomorphisms()
    p.rmsd_filter(0.3)
    assert p.size == 3



import networkx as nx
from networkx.algorithms import isomorphism
import numpy as np
from numpy.linalg import norm

RADII = {
    'H': 0.31,
    'He': 0.28,
    'Li': 1.28,
    'Be': 0.96,
    'B': 0.84,
    'C': 0.69,
    'N': 0.71,
    'O': 0.66,
    'F': 0.57,
    'Ne': 0.58,
    'Na': 1.66,
    'Mg': 1.41,
    'Al': 1.21,
    'Si': 1.11,
    'P': 1.07,
    'S': 1.05,
    'Cl': 1.02,
    'Ar': 1.06,
    'K': 2.03,
    'Ca': 1.76,
    'Sc': 1.7,
    'Ti': 1.6,
    'V': 1.53,
    'Cr': 1.39,
    'Mn': 1.61,
    'Fe': 1.52,
    'Co': 1.5,
    'Ni': 1.24,
    'Cu': 1.32,
    'Zn': 1.22,
    'Ga': 1.22,
    'Ge': 1.2,
    'As': 1.19,
    'Se': 1.2,
    'Br': 1.2,
    'Kr': 1.16,
    'Rb': 2.2,
    'Sr': 1.95,
    'Y': 1.9,
    'Zr': 1.75,
    'Nb': 1.64,
    'Mo': 1.54,
    'Tc': 1.47,
    'Ru': 1.46,
    'Rh': 1.42,
    'Pd': 1.39,
    'Ag': 1.45,
    'Cd': 1.44,
    'In': 1.42,
    'Sn': 1.39,
    'Sb': 1.39,
    'Te': 1.38,
    'I': 1.39,
    'Xe': 1.4,
    'Cs': 2.44,
    'Ba': 2.15,
    'La': 2.07,
    'Ce': 2.04,
    'Pr': 2.03,
    'Nd': 2.01,
    'Pm': 1.99,
    'Sm': 1.98,
    'Eu': 1.98,
    'Gd': 1.96,
    'Tb': 1.94,
    'Dy': 1.92,
    'Ho': 1.92,
    'Er': 1.89,
    'Tm': 1.9,
    'Yb': 1.87,
    'Lu': 1.87,
    'Hf': 1.75,
    'Ta': 1.7,
    'W': 1.62,
    'Re': 1.51,
    'Os': 1.44,
    'Ir': 1.41,
    'Pt': 1.36,
    'Au': 1.36,
    'Hg': 1.32,
    'Tl': 1.45,
    'Pb': 1.46,
    'Bi': 1.48,
    'Po': 1.4,
    'At': 1.5,
    'Rn': 1.5,
    'Fr': 2.6,
    'Ra': 2.21,
    'Ac': 2.15,
    'Th': 2.06,
    'Pa': 2.0,
    'U': 1.96,
    'Np': 1.9,
    'Pu': 1.87,
    'Am': 1.8,
    'Cm': 1.69
}


def write_sdf(G, p, index, fname):
    structure = p[index].xyz
    atom_symbols = p.atom_symbols
    lines = ["", "", ""]
    lines.append("%3d%3d  0  0  0  0  0  0  0  0999 V2000" %
                 (G.number_of_nodes(), G.number_of_edges()))
    #   -5.5250    1.6470    1.0014 C   0  0  0  0  0  0  0  0  0  0  0  0
    new_indexing = list(G.nodes)
    for atom in new_indexing:
        lines.append(
            "%10.4f%10.4f%10.4f%3s  0  0  0  0  0  0  0  0  0  0  0  0" %
            (structure[atom][0], structure[atom][1], structure[atom][2],
             atom_symbols[atom]))

    for edge in G.edges:
        lines.append("%3s%3s%3s  0" % (new_indexing.index(edge[0]) + 1,
                                       new_indexing.index(edge[1]) + 1, 1))
    lines.append("M  END\n")

    with open(fname, "w") as f:
        f.write("\n".join(lines))


def preprocess_kwargs(kwargs):
    default_keys = {
        'ignore_elements': [],
        'mult': 1.0,
        'add_bonds': [],
        'remove_bonds': [],
        'sdf_name': None
    }
    for key in default_keys:
        if key not in kwargs:
            kwargs[key] = default_keys[key]


def extract_subgraph(mygraph, kwargs):
    if len(kwargs['ignore_elements']) > 0 and all(
        [isinstance(i, str) for i in kwargs['ignore_elements']]):
        ignore_mode = 'Elements'
    elif len(kwargs['ignore_elements']) > 0 and all(
        [isinstance(i, int) for i in kwargs['ignore_elements']]):
        ignore_mode = 'Indices'
    else:
        assert len(kwargs['ignore_elements']
                   ) == 0, "Unexpected contents of ignore_elements: " + repr(
                       kwargs['ignore_elements'])
        ignore_mode = 'None'

    if ignore_mode == 'Elements':
        save_atoms = []
        for node in mygraph.nodes:
            if 'HCarbon' in kwargs['ignore_elements'] and mygraph.nodes[node][
                    'symbol'] == 'H':
                nb_list = list(mygraph.neighbors(node))
                if len(nb_list) == 1 and mygraph.nodes[
                        nb_list[0]]['symbol'] == 'C':
                    continue
            if mygraph.nodes[node]['symbol'] not in kwargs['ignore_elements']:
                save_atoms.append(node)
    elif ignore_mode == 'Indices':
        save_atoms = [
            node for node in mygraph.nodes
            if node not in kwargs['ignore_elements']
        ]
    else:  # if ignore_mode == 'None'
        save_atoms = [node for node in mygraph.nodes]

    for bond in kwargs['add_bonds']:
        atA, atB = bond[0] - 1, bond[1] - 1
        assert atA in save_atoms and atB in save_atoms, f"Found a bug: atom(s) {atA}/{atB} not in {repr(save_atoms)}"
        mygraph.add_edge(atA, atB)
    for bond in kwargs['remove_bonds']:
        atA, atB = bond[0] - 1, bond[1] - 1
        assert atA in save_atoms and atB in save_atoms, f"Found a bug: atom(s) {atA}/{atB} not in {repr(save_atoms)}"
        mygraph.remove_edge(atA, atB)

    mysubgraph = mygraph.subgraph(save_atoms)
    return mysubgraph


def generate_connectivity(p, index, kwargs):
    preprocess_kwargs(kwargs)

    mygraph = nx.Graph()
    structure = p[index].xyz
    atom_symbols = p.atom_symbols

    for i, symbol in enumerate(atom_symbols):
        # if symbol not in ignore_elements:
        mygraph.add_node(i)
        mygraph.nodes[i]['symbol'] = symbol
    for nodeA in range(len(atom_symbols)):
        for nodeB in range(nodeA):
            max_dist = kwargs['mult'] * (RADII[atom_symbols[nodeA]] +
                                         RADII[atom_symbols[nodeB]])
            if norm(structure[nodeA] - structure[nodeB]) < max_dist:
                mygraph.add_edge(nodeA, nodeB)

    mysubgraph = extract_subgraph(mygraph, kwargs)

    if kwargs['sdf_name'] is not None:
        write_sdf(mysubgraph, p, index, kwargs['sdf_name'])
    return mysubgraph


def generate_connectivity_from_graph(p, molgr, kwargs):
    preprocess_kwargs(kwargs)

    mygraph = nx.Graph()
    atom_symbols = p.atom_symbols

    for node, symbol in zip(range(molgr.number_of_nodes()), atom_symbols):
        mygraph.add_node(node)
        mygraph.nodes[node]['symbol'] = symbol
    mygraph.add_edges_from(molgr.edges(data=True))

    mysubgraph = extract_subgraph(mygraph, kwargs)
    return mysubgraph


def same_element(n1_attrib, n2_attrib):
    return n1_attrib['symbol'] == n2_attrib['symbol']


def generate_isomorphisms(mysubgraph):
    GM = isomorphism.GraphMatcher(mysubgraph,
                                  mysubgraph,
                                  node_match=same_element)
    isomorphisms = []
    atoms = list(mysubgraph.nodes)
    for isom in GM.isomorphisms_iter():
        isomorphisms.append([isom[i] for i in atoms])
    return isomorphisms



import copy
import math
import json
import os
import inspect
import networkx as nx
from ringo import base

Problem = base.Problem
# from ringo_base import Problem#, log, log_rstart, log_rend

CACHE_FILE = 'cached_buildseqs.json'


## GENERATE IG
def calc_ndof(G):
    # addconstr = 0
    ncycles = nx.number_of_edges(G) - nx.number_of_nodes(G) + 1
    nnodes = G.number_of_nodes()
    return nnodes - 3 - 3 * ncycles  # - addconstr


def generate_ig(G):
    first_mcb = [set(c) for c in nx.minimum_cycle_basis(G)]
    templist = []
    for item in first_mcb:
        templist.append(sorted(list(item)))
    templist.sort()
    first_mcb = [set(i) for i in templist]

    tempG = nx.Graph()
    for i in range(len(first_mcb)):
        tempG.add_node(i)
        tempG.nodes[i]['atomset'] = first_mcb[i]
        if len(first_mcb[i]) <= 5:
            tempG.nodes[i]['ndof'] = -1  # No DOFs a priory
        else:
            tempG.nodes[i]['ndof'] = len(first_mcb[i]) - 6
    for i in range(tempG.number_of_nodes()):
        for j in range(i + 1, tempG.number_of_nodes()):
            if len(tempG.nodes[i]['atomset'].intersection(
                    tempG.nodes[j]['atomset'])) > 0:
                tempG.add_edge(i, j)

    # Unite all a priory rigid rings
    clean_run = False
    while not clean_run:
        clean_run = True
        start_tempG_nodes = list(tempG.nodes)
        for node in start_tempG_nodes:
            if tempG.nodes[node]['ndof'] != -1:
                continue
            # If the part is rigid (ndof == -1)
            for nbnode in list(tempG.neighbors(node)):
                spiro_link = (len(tempG.nodes[node]['atomset'].intersection(
                    tempG.nodes[nbnode]['atomset'])) == 1)
                lack_dofs = (len(tempG.nodes[nbnode]['atomset']) - 6 -
                             (len(tempG.nodes[node]['atomset'].intersection(
                                 tempG.nodes[nbnode]['atomset'])) - 1) < 0)

                if not spiro_link and lack_dofs:
                    # Should unite 'node' with 'nbnode'
                    tempG.nodes[node]['atomset'].update(
                        tempG.nodes[nbnode]['atomset'])
                    for other_nb in tempG.neighbors(nbnode):
                        if other_nb != node:
                            tempG.add_edge(other_nb, node)
                    tempG.remove_node(nbnode)
                    clean_run = False
            if not clean_run:
                break

    mcb = []
    for node in tempG.nodes():
        mcb.append(tempG.nodes[node]['atomset'])
    IG = nx.Graph()
    for i in range(len(mcb)):
        for j in range(i + 1, len(mcb)):
            common_atoms = mcb[i].intersection(mcb[j])
            if len(common_atoms) > 0:
                IG.add_edge(i, j)
                IG[i][j]['link'] = list(common_atoms)
    if IG.number_of_edges() == 0 and len(mcb) == 1:
        IG.add_node(0)
    assert IG.number_of_nodes() != 0, "Cycle graph has 0 nodes"

    for i in range(len(mcb)):
        fraggraph = nx.Graph()
        for edge in list(G.edges):
            if edge[0] in mcb[i] and edge[1] in mcb[i]:
                issingle = (G[edge[0]][edge[1]]['type'] == 1)
                fraggraph.add_edge(*edge, single=issingle)
        IG.nodes[i]['graph'] = fraggraph
        IG.nodes[i]['NDOF'] = calc_ndof(fraggraph)

    for nA, nB in IG.edges:
        comm_bonds = []
        for edge in IG.nodes[nA]['graph'].edges:
            if IG.nodes[nB]['graph'].has_edge(*edge):
                comm_bonds.append(sorted(list(edge)))
        IG[nA][nB]['common_bonds'] = sorted(comm_bonds)
    return IG


## GENERATE BUILD SEQUENCE
def log_problem(mygraph, consbonds, seq):
    log_bondtypes = []
    for edge in mygraph.edges:
        if mygraph[edge[0]][edge[1]]['type']:
            value = 1
        else:
            value = 0
        log_bondtypes.append([*edge, value])

    cb_repr = None
    if len(consbonds) == 0:
        cb_repr = repr([])
    elif isinstance(consbonds[0], list):
        cb_repr = repr(sorted(consbonds))
    # if cb_repr is not None:
    #     log_rstart("problem_init")
    #     log(" IN:: nodes = " + repr(sorted(list(mygraph.nodes))))
    #     log(" IN:: consbonds = " + cb_repr)
    #     log(" IN:: seq = " + repr(seq))
    #     log(" CHECK:: edges_SET2 = " + repr(list(mygraph.edges)))
    #     log(" CHECK:: bondtypes_SET2 = " + repr(log_bondtypes))
    #     log_rend("problem_init")


def sync_types(IG, node):
    for nb in IG.neighbors(node):
        link = IG[node][nb]['link']
        if len(link) == 2:
            if not IG.nodes[node]['graph'][link[0]][link[1]]['type']:
                IG.nodes[nb]['graph'][link[0]][link[1]]['type'] = False
        elif len(link) > 2:
            for edge in IG.nodes[node]['graph'].edges():
                if edge[0] in link and edge[1] in link and \
                        not IG.nodes[node]['graph'][edge[0]][edge[1]]['type']:
                    IG.nodes[nb]['graph'][edge[0]][edge[1]]['type'] = False
    for nb in IG.neighbors(node):
        link = IG[node][nb]['link']
        if len(link) == 2:
            if not IG.nodes[node]['graph'][link[0]][link[1]]['unrestrained']:
                IG.nodes[nb]['graph'][link[0]][link[1]]['unrestrained'] = False
        elif len(link) > 2:
            for edge in IG.nodes[node]['graph'].edges():
                if edge[0] in link and edge[1] in link and \
                        not IG.nodes[node]['graph'][edge[0]][edge[1]]['unrestrained']:
                    IG.nodes[nb]['graph'][edge[0]][
                        edge[1]]['unrestrained'] = False


def process_seq(seq, IG, requested_free, final=True):
    found_bonds = [False] * len(requested_free)
    for node in seq:
        G = IG.nodes[node]['graph']
        for edge in G.edges:
            G[edge[0]][edge[1]]['type'] = G[edge[0]][edge[1]]['single']
            G[edge[0]][edge[1]]['unrestrained'] = True
        for i, bond in enumerate(requested_free):
            if G.has_edge(*bond):
                found_bonds[i] = True
    assert False not in found_bonds

    LG = nx.DiGraph(directed=True)
    for node in seq:
        LG.add_node(node)
        for nb in IG.neighbors(node):
            if LG.has_node(nb):
                LG.add_edge(nb, node)

    for node in LG.nodes:
        nb_list = list(LG.neighbors(node))
        LG.nodes[node]['dep_max'] = len(nb_list)
        LG.nodes[node]['dep_cur'] = 0
        LG.nodes[node]['ikprob'] = None

    LG = LG.reverse(copy=False)
    for node in LG.nodes:
        nb_list = list(LG.neighbors(node))
        lbs = []
        for nb in nb_list:
            for bond in IG[node][nb]['common_bonds']:
                if bond not in lbs:  # and [bond[1], bond[0]] not in lbs:
                    lbs.append(bond)
        LG.nodes[node]['linking_bonds'] = lbs

    done = False
    onemorerun = True

    reached_ends = 0
    num_of_ends = len(
        [node for node in LG.nodes if len(list(LG.neighbors(node))) == 0])
    while not done:
        for node in seq:
            if LG.nodes[node]['dep_cur'] == LG.nodes[node]['dep_max']:
                if LG.nodes[node]['ikprob'] is None:
                    current_request_free = []
                    for bond in requested_free:
                        if IG.nodes[node]['graph'].has_edge(*bond) and \
                           list(bond) not in LG.nodes[node]['linking_bonds'] and \
                           [bond[1], bond[0]] not in LG.nodes[node]['linking_bonds']:
                            current_request_free.append(bond)
                    LG.nodes[node]['ikprob'] = Problem(
                        IG.nodes[node]['graph'],
                        LG.nodes[node]['linking_bonds'], current_request_free,
                        final)
                    sync_types(IG, node)
                elif not LG.nodes[node]['ikprob'].recheck_method():
                    sync_types(IG, node)
                    onemorerun = True

                if len(list(LG.neighbors(node))) == 0:
                    reached_ends += 1

                if reached_ends == num_of_ends:
                    reached_ends = 0
                    done = True
                    if onemorerun:
                        onemorerun = False
                        done = False
                        for node in LG.nodes():
                            LG.nodes[node]['dep_cur'] = 0
                    continue

                for nb in LG.neighbors(node):
                    LG.nodes[nb]['dep_cur'] += 1
                LG.nodes[node][
                    'dep_cur'] += 1  # To avoid repetitions for the same node

    if final:
        # We have TLC applicability warnings prepared during iterations of 'while'
        # but they are recored afterwards.
        for node in seq:
            LG.nodes[node]['ikprob'].record_warning()

        blocked_graph = nx.DiGraph(directed=True)
        blocked_graph.add_edges_from(LG.edges)
        nonblocked_nodes = []
        for node in blocked_graph.nodes:
            frag_graph = IG.nodes[node]['graph']

            # Not interested in rings that are processed by TLC
            unrestrained = False
            for bondA, bondB in frag_graph.edges:
                if frag_graph[bondA][bondB]['unrestrained']:
                    unrestrained = True
                    break
            assert ((LG.nodes[node]['ikprob'].method == 1) and unrestrained) or \
                ((LG.nodes[node]['ikprob'].method != 1) and not unrestrained)
            if unrestrained:
                nonblocked_nodes.append(node)
                continue

            # Not interested in rings that are rigid purely by their NDOFs
            nedges = frag_graph.number_of_edges()
            nnodes = frag_graph.number_of_nodes()
            nrings = nedges - nnodes + 1
            nfixed = 0
            for edge in frag_graph.edges:
                if not frag_graph[edge[0]][edge[1]]['single']:
                    nfixed += 1
            if nedges - nrings * 6 - nfixed < 0:
                # print(f"Node #{node} is rigid purely by NDOFs. nedges={nedges} nnodes={nnodes} nrings={nrings} nfixed={nfixed}")
                nonblocked_nodes.append(node)

        # Spiro-linkages do not create any dependencies between rings
        # Removing them will make generated warnings clearer
        remove_edges = []
        for nodeA, nodeB in blocked_graph.edges:
            if len(IG[nodeA][nodeB]['common_bonds']) == 0:
                remove_edges.append((nodeA, nodeB))
        blocked_graph.remove_edges_from(remove_edges)

        # Remove the fragments that are totally not flexible or IK will be applied
        blocked_graph.remove_nodes_from(nonblocked_nodes)

        if blocked_graph.number_of_nodes() > 0:
            blocked_undirected = nx.Graph()
            blocked_undirected.add_edges_from(blocked_graph.edges)

            # Each connected component represents a potentially flexible system
            # that falls under limitations for our polycycle processing method
            for component_nodes in nx.connected_components(blocked_undirected):
                component_graph = IG.subgraph(component_nodes)
                component_moledges = []
                for node in component_graph.nodes:
                    component_moledges += IG.nodes[node]['graph'].edges

                component_molgraph = nx.Graph()
                component_molgraph.add_edges_from(component_moledges)
                for node in component_graph.nodes:
                    for edge in IG.nodes[node]['graph'].edges:
                        component_molgraph[edge[0]][
                            edge[1]]['single'] = IG.nodes[node]['graph'][
                                edge[0]][edge[1]]['single']

                nedges = component_molgraph.number_of_edges()
                nnodes = component_molgraph.number_of_nodes()
                nrings = nedges - nnodes + 1
                nfixed = 0
                for edge in component_molgraph.edges:
                    if not component_molgraph[edge[0]][edge[1]]['single']:
                        nfixed += 1

                if nedges - nrings * 6 - nfixed >= 0:
                    message_data = {
                        "message":
                        f"Unable to sample kinematically flexible {len(component_nodes)}-cyclic fragment ({component_molgraph.number_of_nodes()} atoms in the fragment). This fragment will be kept as it is in the starting conformation.",
                        "subject":
                        Problem.get_warning_codes()["IK_NOT_APPLIED"] +
                        "[important]",
                        "atoms": list(component_molgraph.nodes),
                        "file": __file__,
                        "line": inspect.currentframe().f_lineno,
                    }
                    Problem.add_message_to_feed(json.dumps(message_data))
                else:
                    message_data = {
                        "message":
                        f"Encountered kinematically rigid polycycle. Can continue. {len(component_nodes)}-cyclic fragment ({component_molgraph.number_of_nodes()} atoms in the fragment).",
                        "subject":
                        Problem.get_warning_codes()
                        ["RIGIDITY_ANALYSIS_UNRELIABLE"] + "[important]",
                        "atoms":
                        list(component_molgraph.nodes),
                        "file":
                        __file__,
                        "line":
                        inspect.currentframe().f_lineno,
                    }
                    Problem.add_message_to_feed(json.dumps(message_data))

        # Add warnings if there are unfulfilled DOF requests
        for node in LG.nodes:
            unfulfilled = LG.nodes[node]['ikprob'].get_unfulfilled_requests()
            for bond in unfulfilled:
                message_data = {
                    "message":
                    f"Unable to enforce the bond {[i+1 for i in bond]} to be a free DOF of the ring(s) {repr([i+1 for i in IG.nodes[node]['graph'].nodes])}",
                    "subject":
                    Problem.get_warning_codes()["UNMET_DOF_REQUEST"] +
                    "[important]",
                    "atoms": list(bond),
                    "file": __file__,
                    "line": inspect.currentframe().f_lineno,
                }
                Problem.add_message_to_feed(json.dumps(message_data))
    return LG


def strats_with_headnode(headnode, IG, requested_free):
    done_seqs = []
    prospects = [[headnode]]
    while len(prospects) > 0:
        cur_prospect = prospects.pop()
        # TODO Check cur_prospect is okay

        if len(cur_prospect) < IG.number_of_nodes():
            for node in IG.nodes:
                if node not in cur_prospect:
                    prospects.append(cur_prospect + [node])
        elif cur_prospect not in done_seqs:
            done_seqs.append(cur_prospect)

    scores = []
    # for node in IG.nodes:
    #     print(f"Nodes of CP#{node} = {repr(list(IG.nodes[node]['graph'].nodes))}")
    # print(f"SEQS = {repr(done_seqs)}")
    for seq_idx, seq in enumerate(done_seqs):
        # print(f"----------- SEQ #{seq_idx} -----------")
        LG = process_seq(seq, IG, requested_free, final=False)

        score = 0.0
        for node in LG.nodes:
            if LG.nodes[node]['ikprob'].method != 0:
                score += 1.0
            unfulfilled = LG.nodes[node]['ikprob'].get_unfulfilled_requests()
            score -= len(unfulfilled) * 0.01
        scores.append(score)

    best_score = max(scores)
    best_strats = [
        done_seqs[i] for i in range(len(scores)) if scores[i] == best_score
    ]

    log_atoms = set()
    for node in IG.nodes:
        G = IG.nodes[node]['graph']
        log_atoms.update(set(G.nodes))
    # log_rstart("stratheadnode_final")
    # log(" IN:: nodes = " + repr(sorted(list(IG.nodes))))
    # log(" IN:: headnode = " + repr(headnode))
    # log(" IN:: atoms = " + repr(sorted(list(log_atoms))))
    # log(" CHECK:: headatoms = " + repr(sorted(list(IG.nodes[headnode]['graph'].nodes))))
    # log(" CHECK:: bestscore = " + repr([best_score]))
    # log(" CHECK:: best_strats = " + repr(best_strats))
    # log_rend("stratheadnode_final")

    return best_strats, best_score


def get_strategy_simple(IG):
    LG = nx.Graph()
    max_dof_value = None
    max_dof_node = None
    for node in IG.nodes:
        LG.add_node(node)
        if max_dof_value is None or max_dof_value < IG.nodes[node]['NDOF']:
            max_dof_value = IG.nodes[node]['NDOF']
            max_dof_node = node

    done = False
    onemorerun = True
    seq = [max_dof_node]  # Built in reverse order
    while not done:
        done = True
        for node in seq:
            for nb in IG.neighbors(node):
                if nb not in seq:
                    LG.add_edge(node, nb)
                    seq.append(nb)
                    onemorerun = True
                    done = False
        if onemorerun:
            done = False
            onemorerun = False
    seq = [i for i in reversed(seq)]  # Assemble in normal order
    return seq


def get_strategy(IG, filename, requested_free, require_best_sequence):
    # Don't waste time in the most trivial cases
    if IG.number_of_nodes() == 1:
        assert IG.has_node(0)
        return [0]

    cached_data = {}
    if os.path.exists(CACHE_FILE):
        with open(CACHE_FILE, 'r') as f:
            cached_data = json.load(f)

    ig_data = {
        'ignodes':
        sorted(list(IG.nodes)),
        'igedges':
        sorted([sorted(list(edge)) for edge in IG.edges]),
        'cycles':
        [sorted(list(IG.nodes[node]['graph'].nodes)) for node in IG.nodes]
    }

    # Check if the current cyclic part is cached in the file
    present = False
    cached_index = None
    if filename in cached_data:
        for cur_index, cur_data in enumerate(cached_data[filename]):
            if cur_data['id'] == ig_data:
                present = True
                cached_index = cur_index

    # Get build seq from cached_data
    if present:
        build_seq = cached_data[filename][cached_index]['seq']
        # print(f'Build sequence has been retrieved from {CACHE_FILE}')
        return build_seq

    # If cache is not available, compute build seq from scratch
    # print(f'Number of required iterations = {math.factorial(IG.number_of_nodes())}')
    if not require_best_sequence and math.factorial(
            IG.number_of_nodes()) > 720:
        # Post a warning that bruteforce is too demanding
        atoms_set = set()
        for node in IG.nodes:
            G = IG.nodes[node]['graph']
            atoms_set.update(set(G.nodes))
        add_warning = ""
        if len(requested_free) > 0:
            add_warning = " Unable to enforce the requested dihedral(s)."
        message_data = {
            "message":
            f"Cannot use bruteforce to find the best solution sequence. If there are no further warnings IK will still be applicable.{add_warning} In case if this causes any problems consider using require_best_sequence=True.",
            "subject": Problem.get_warning_codes()["SUBOPTIMAL_SOLN_SEQ"],
            "atoms": sorted(list(atoms_set)),
            "file": __file__,
            "line": inspect.currentframe().f_lineno,
        }
        Problem.add_message_to_feed(json.dumps(message_data))

        res = get_strategy_simple(IG)
    else:
        # If bruteforce is feasible
        best_strat = None
        best_score = None

        for headnode in IG.nodes:
            strats, score = strats_with_headnode(headnode, IG, requested_free)
            if best_score is None or best_score < score:
                best_score = score
                best_strat = strats[0]

        log_atoms = set()
        for node in IG.nodes:
            G = IG.nodes[node]['graph']
            log_atoms.update(set(G.nodes))
        res = best_strat

    # Cache the build seq
    if filename not in cached_data:
        cached_data[filename] = []
    cached_data[filename].append({
        'id': ig_data,
        'seq': res,
    })
    with open(CACHE_FILE, 'w') as f:
        json.dump(cached_data, f)
    return res



import os
from typing import Any, List

curdir = os.path.dirname(os.path.realpath(__file__))

DOUBLE_THR = 1e-6


def to_abs(path):
    return os.path.join(curdir, path)


def is_float(inp: Any) -> bool:
    if isinstance(inp, str):
        is_float = False
        try:
            float(inp)
            is_float = True
        except:
            pass
        return is_float
    else:
        return isinstance(inp, float) or isinstance(inp, int)


def is_int(inp: Any) -> bool:
    if isinstance(inp, str):
        is_int = False
        try:
            int(inp)
            is_int = True
        except:
            pass
        return is_int
    else:
        return isinstance(inp, int)


def match_lists(a: List, b: List) -> bool:
    # print("{} vs. {}".format(repr(a), repr(b)))
    if len(a) != len(b):
        return False

    if len(a) == 0:
        print("Len {} vs. {}".format(repr(a), repr(b)))
        return True
    okay = True
    for item_a, item_b in zip(a, b):
        if isinstance(item_a, list) and isinstance(item_b, list):
            # if not match_lists(item_a, item_b):
            #     print("Lists {} vs. {}".format(repr(item_a), repr(item_b)))
            okay = okay and match_lists(item_a, item_b)
        elif is_int(item_a) and is_int(item_b):
            # if not a == b:
            #     print("Ints {} vs. {}".format(repr(item_a), repr(item_b)))
            okay = okay and (a == b)
        elif is_float(item_a) and is_float(item_b):
            if not (abs(item_a - item_b) < DOUBLE_THR):
                print("Floats {} vs. {}".format(repr(item_a), repr(item_b)))
            okay = okay and (abs(item_a - item_b) < DOUBLE_THR)
        else:
            raise RuntimeError("Unexpected types of {} and {}".format(
                repr(a), repr(b)))
    return okay



import os
import sys

MAIN_PACKAGE = False

if MAIN_PACKAGE:
    sys.path.insert(0, os.path.dirname(__file__))
    from .cpppart import cpppart as base
else:
    from ..cpppart import cpppart as base


def _rmsd_interval(self, start_rmsd, end_rmsd, matr):
    min_rmsd = min(start_rmsd, end_rmsd)
    max_rmsd = max(start_rmsd, end_rmsd)
    ascending = 1 if start_rmsd < end_rmsd else -1
    assert matr.ndim == 2
    assert matr.shape[0] == len(self) and matr.shape[1] == len(self)

    df = {'molA': [], 'molB': [], 'rmsd': []}
    for i in range(matr.shape[0]):
        for j in range(i):
            if matr[i, j] > min_rmsd and matr[i, j] < max_rmsd:
                df['molA'].append(i)
                df['molB'].append(j)
                df['rmsd'].append(matr[i, j])

    df['molA'], df['molB'], df['rmsd'] = zip(
        *sorted(zip(df['molA'], df['molB'], df['rmsd']),
                key=lambda x: ascending * x[2]))

    for indexA, indexB, rmsd in zip(df['molA'], df['molB'], df['rmsd']):
        yield self[indexA], self[indexB], float(rmsd)


base.Confpool.rmsd_fromto = _rmsd_interval



import platform
import glob
import ntpath
import os
from setuptools import setup

# OS specifics
CUR_OS = platform.system()
SHAREDOBJ_TEMPLATE = {
    'Windows': "ringo_base.cp{py_ver}-win_amd64.pyd",
    'Linux': "ringo_base.cpython-{py_ver}*-x86_64-linux-gnu.so",
}

# assert CUR_OS in ['Linux',
#                   'Windows'], "Only Linux and Windows platforms are supported"
assert CUR_OS == 'Linux', "Only Linux platform is currently supported 🤗"

# Python version specifics
python_version_tuple = platform.python_version_tuple()
py_ver = int(f"{python_version_tuple[0]}{python_version_tuple[1]}")

ringo_so_list = glob.glob(
    os.path.join('./ringo', SHAREDOBJ_TEMPLATE[CUR_OS].format(py_ver=py_ver)))
assert len(ringo_so_list) == 1
ringo_object_name = ntpath.basename(ringo_so_list[0])

for file in glob.glob('./ringo/*.pyd') + glob.glob('./ringo/*.so'):
    if ntpath.basename(file) != ringo_object_name:
        os.remove(file)

if CUR_OS == 'Windows':
    ADDITIONAL_FILES = ['*.dll']
elif CUR_OS == 'Linux':
    ADDITIONAL_FILES = []


def all_py(dir='.', ext='.py'):
    result = []
    for root, _, files in os.walk(dir):
        for file in files:
            if file.endswith(ext):
                result.append(
                    os.path.join(
                        *(os.path.join(root, file).split(os.sep)[1:])))
    return result


setup(
    name='ringo_ik',
    version='1.0.1',
    author='Nikolai Krivoshchapov',
    python_requires=f'=={python_version_tuple[0]}.{python_version_tuple[1]}.*',
    install_requires=[
        'numpy',
        'networkx',
    ],
    platforms=['Windows', 'Linux'],
    packages=['ringo'],
    package_data={
        'ringo': [
            ringo_object_name,
            *all_py('ringo'),  # All python files
            *all_py('ringo', ext='.xyz'),  # Structures for built-in testing
            *all_py('ringo', ext='.sdf'),  # Structures for built-in testing
            *ADDITIONAL_FILES,
        ]
    })



import random
import time
import math
import os
import multiprocessing

from functools import wraps
from typing import Callable

import numpy as np

from .utils import (
    match_lists,
    to_abs,
    DOUBLE_THR,
)

from .. import (
    Molecule,
    Confpool,
    base,
    run_confsearch,
    get_molecule_statistics,
)

if base.use_pyxyz:
    from .. import Confpool

RINGO_TESTS = {}


def test(raw_f: Callable, name=None) -> Callable:

    @wraps(raw_f)
    def f():
        try:
            raw_f()
            success = True
            message = ''
        except Exception as e:
            success = False
            message = repr(e)
        return (success, message)

    RINGO_TESTS[f.__name__ if name is None else name] = f
    return f


@test
def manual_dofs():
    mol = Molecule(sdf=to_abs('pdb1GHG.sdf'))
    dofs_list, dofs_values = mol.get_ps()
    ddofs_list, ddofs_values = mol.get_discrete_ps()
    while True:
        for i in range(len(dofs_list)):
            newvalue = random.uniform(-math.pi, math.pi)
            dofs_values[i] = newvalue

        for i in range(len(ddofs_list)):
            ddofs_values[i] = -1  # -1 requests for random solution of IK

        result = mol.apply_ps()

        if result == 0:
            symbols = mol.get_symbols()
            coords = mol.get_xyz()
            assert len(symbols) == coords.shape[0]
            break


@test
def mol_stats():
    mol = Molecule(sdf=to_abs('pdb1GHG.sdf'))
    res = get_molecule_statistics(mol)
    assert all(key in res
               for key in ('composition', 'num_atoms', 'num_heavy_atoms',
                           'num_bonds', 'num_rotatable_bonds',
                           'num_cyclic_rotatable_bonds',
                           'largest_macrocycle_size', 'n_flexible_rings',
                           'n_rigid_rings', 'num_dofs', 'num_cyclic_dofs',
                           'cyclomatic_number', 'num_cyclic_parts'))


@test
def all_solutions():
    mol = Molecule(sdf=to_abs('pdb1GHG.sdf'))
    dofs_list, dofs_values = mol.get_ps()
    ddofs_list, ddofs_values = mol.get_discrete_ps()
    while True:
        for i in range(len(dofs_list)):
            newvalue = random.uniform(-math.pi, math.pi)
            dofs_values[i] = newvalue

        result = mol.prepare_solution_iterator()
        if result != 0:
            continue

        sol_list = mol.get_solutions_list()
        if len(sol_list) > 2:
            break


def gen_sampling(parallel: bool):

    def f():
        if parallel:
            assert base.use_mcr_parallel and base.use_pyxyz, "PyXYZ or MCR are disabled in this build"
        else:
            assert base.use_mcr_singlethreaded and base.use_pyxyz, "PyXYZ or MCR are disabled in this build"
        mol = Molecule(sdf=to_abs('pdb1GHG.sdf'))
        p = Confpool()

        nthreads = 1 if not parallel else min(multiprocessing.cpu_count(), 8)
        run_confsearch(
            mol,
            pool=p,
            postopt_settings=[{
                'enabled': False
            }, {
                'enabled': False
            }],
            geometry_validation={
                "ringo": {
                    "bondlength": 0.05,
                    "valence": 3.0,
                    "dihedral": 3.0,
                }
            },
            timelimit=10,
            rmsd_settings='default',
            nthreads=nthreads,
        )
        assert p.size > 0, "No conformers, that's sus"

    return f


test(gen_sampling(parallel=False), name='sampling_singlethread')
test(gen_sampling(parallel=True), name='sampling_parallel')



from .base_load import base
from .molgraph import Molecule

DEG2RAD = 0.0174532925199432957692
RAD2DEG = 1 / DEG2RAD
H2KC = 627.509474063
KC2H = 1 / H2KC
Confpool = base.Confpool
MolProxy = base.MolProxy
